/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.github.lgooddatepicker.components;

/**
 *
 * @author mcwolfy
 */
public interface ComponentListener {
	public void componentCreated(ComponentEvent e);
}
